@@ -3,51 +3,55 @@ from __future__ import annotations
 import typing as t
 
 from sqlglot import exp, transforms
+from sqlglot.dialects.dialect import concat_to_dpipe_sql, rename_func
 from sqlglot.dialects.postgres import Postgres
 from sqlglot.helper import seq_get
 from sqlglot.tokens import TokenType
 
 
-def _json_sql(self, e) -> str:
-    return f'{self.sql(e, "this")}."{e.expression.name}"'
+def _json_sql(self: Postgres.Generator, expression: exp.JSONExtract | exp.JSONExtractScalar) -> str:
+    return f'{self.sql(expression, "this")}."{expression.expression.name}"'
 
 
 class Redshift(Postgres):
-    time_format = "'YYYY-MM-DD HH:MI:SS'"
-    time_mapping = {
-        **Postgres.time_mapping,  # type: ignore
+    TIME_FORMAT = "'YYYY-MM-DD HH:MI:SS'"
+    TIME_MAPPING = {
+        **Postgres.TIME_MAPPING,
         "MON": "%b",
         "HH": "%H",
     }
 
     class Parser(Postgres.Parser):
         FUNCTIONS = {
-            **Postgres.Parser.FUNCTIONS,  # type: ignore
+            **Postgres.Parser.FUNCTIONS,
             "DATEADD": lambda args: exp.DateAdd(
-                this=seq_get(args, 2),
+                this=exp.TsOrDsToDate(this=seq_get(args, 2)),
                 expression=seq_get(args, 1),
                 unit=seq_get(args, 0),
             ),
             "DATEDIFF": lambda args: exp.DateDiff(
-                this=seq_get(args, 2),
-                expression=seq_get(args, 1),
+                this=exp.TsOrDsToDate(this=seq_get(args, 2)),
+                expression=exp.TsOrDsToDate(this=seq_get(args, 1)),
                 unit=seq_get(args, 0),
             ),
             "NVL": exp.Coalesce.from_arg_list,
+            "STRTOL": exp.FromBase.from_arg_list,
         }
 
         CONVERT_TYPE_FIRST = True
 
-        def _parse_types(self, check_func: bool = False) -> t.Optional[exp.Expression]:
-            this = super()._parse_types(check_func=check_func)
+        def _parse_types(
+            self, check_func: bool = False, schema: bool = False
+        ) -> t.Optional[exp.Expression]:
+            this = super()._parse_types(check_func=check_func, schema=schema)
 
             if (
                 isinstance(this, exp.DataType)
-                and this.this == exp.DataType.Type.VARCHAR
+                and this.is_type("varchar")
                 and this.expressions
                 and this.expressions[0].this == exp.column("MAX")
             ):
-                this.set("expressions", [exp.Var(this="MAX")])
+                this.set("expressions", [exp.var("MAX")])
 
             return this
 
@@ -57,7 +61,7 @@ class Redshift(Postgres):
         STRING_ESCAPES = ["\\"]
 
         KEYWORDS = {
-            **Postgres.Tokenizer.KEYWORDS,  # type: ignore
+            **Postgres.Tokenizer.KEYWORDS,
             "HLLSKETCH": TokenType.HLLSKETCH,
             "SUPER": TokenType.SUPER,
             "SYSDATE": TokenType.CURRENT_TIMESTAMP,
@@ -77,19 +81,20 @@ class Redshift(Postgres):
         RENAME_TABLE_WITH_DB = False
 
         TYPE_MAPPING = {
-            **Postgres.Generator.TYPE_MAPPING,  # type: ignore
+            **Postgres.Generator.TYPE_MAPPING,
             exp.DataType.Type.BINARY: "VARBYTE",
             exp.DataType.Type.VARBINARY: "VARBYTE",
             exp.DataType.Type.INT: "INTEGER",
         }
 
         PROPERTIES_LOCATION = {
-            **Postgres.Generator.PROPERTIES_LOCATION,  # type: ignore
+            **Postgres.Generator.PROPERTIES_LOCATION,
             exp.LikeProperty: exp.Properties.Location.POST_WITH,
         }
 
         TRANSFORMS = {
-            **Postgres.Generator.TRANSFORMS,  # type: ignore
+            **Postgres.Generator.TRANSFORMS,
+            exp.Concat: concat_to_dpipe_sql,
             exp.CurrentTimestamp: lambda self, e: "SYSDATE",
             exp.DateAdd: lambda self, e: self.func(
                 "DATEADD", exp.var(e.text("unit") or "day"), e.expression, e.this
@@ -99,10 +104,13 @@ class Redshift(Postgres):
             ),
             exp.DistKeyProperty: lambda self, e: f"DISTKEY({e.name})",
             exp.DistStyleProperty: lambda self, e: self.naked_property(e),
+            exp.FromBase: rename_func("STRTOL"),
             exp.JSONExtract: _json_sql,
             exp.JSONExtractScalar: _json_sql,
+            exp.SafeConcat: concat_to_dpipe_sql,
             exp.Select: transforms.preprocess([transforms.eliminate_distinct_on]),
             exp.SortKeyProperty: lambda self, e: f"{'COMPOUND ' if e.args['compound'] else ''}SORTKEY({self.format_args(*e.this)})",
+            exp.TsOrDsToDate: lambda self, e: self.sql(e.this),
         }
 
         # Postgres maps exp.Pivot to no_pivot_sql, but Redshift support pivots
@@ -111,7 +119,7 @@ class Redshift(Postgres):
         # Redshift uses the POW | POWER (expr1, expr2) syntax instead of expr1 ^ expr2 (postgres)
         TRANSFORMS.pop(exp.Pow)
 
-        RESERVED_KEYWORDS = {*Postgres.Generator.RESERVED_KEYWORDS, "snapshot"}
+        RESERVED_KEYWORDS = {*Postgres.Generator.RESERVED_KEYWORDS, "snapshot", "type"}
 
         def values_sql(self, expression: exp.Values) -> str:
             """
@@ -140,7 +148,7 @@ class Redshift(Postgres):
 
                 selects.append(exp.Select(expressions=row))
 
-            subquery_expression = selects[0]
+            subquery_expression: exp.Select | exp.Union = selects[0]
             if len(selects) > 1:
                 for select in selects[1:]:
                     subquery_expression = exp.union(subquery_expression, select, distinct=False)
@@ -158,12 +166,12 @@ class Redshift(Postgres):
             without precision we convert it to `VARCHAR(max)` and if it does have precision then we just convert
             `TEXT` to `VARCHAR`.
             """
-            if expression.this == exp.DataType.Type.TEXT:
+            if expression.is_type("text"):
                 expression = expression.copy()
                 expression.set("this", exp.DataType.Type.VARCHAR)
                 precision = expression.args.get("expressions")
 
                 if not precision:
-                    expression.append("expressions", exp.Var(this="MAX"))
+                    expression.append("expressions", exp.var("MAX"))
 
             return super().datatype_sql(expression)
