@@ -4,15 +4,17 @@ import itertools
 import typing as t
 
 from sqlglot import alias, exp
+from sqlglot._typing import E
+from sqlglot.dialects.dialect import DialectType
 from sqlglot.errors import OptimizeError
-from sqlglot.helper import seq_get
+from sqlglot.helper import case_sensitive, seq_get
 from sqlglot.optimizer.scope import Scope, traverse_scope, walk_in_scope
 from sqlglot.schema import Schema, ensure_schema
 
 
 def qualify_columns(
     expression: exp.Expression,
-    schema: dict | Schema,
+    schema: t.Dict | Schema,
     expand_alias_refs: bool = True,
     infer_schema: t.Optional[bool] = None,
 ) -> exp.Expression:
@@ -91,7 +93,7 @@ def _pop_table_column_aliases(derived_tables):
 
 def _expand_using(scope, resolver):
     joins = list(scope.find_all(exp.Join))
-    names = {join.this.alias for join in joins}
+    names = {join.this.alias_or_name for join in joins}
     ordered = [key for key in scope.selected_sources if key not in names]
 
     # Mapping of automatically joined column names to an ordered set of source names (dict).
@@ -404,9 +406,6 @@ def _qualify_outputs(scope):
             selection = alias(
                 selection,
                 alias=selection.output_name or f"_col_{i}",
-                quoted=True
-                if isinstance(selection, exp.Column) and selection.this.quoted
-                else None,
             )
         if aliased_column:
             selection.set("alias", exp.to_identifier(aliased_column))
@@ -416,6 +415,23 @@ def _qualify_outputs(scope):
     scope.expression.set("expressions", new_selections)
 
 
+def quote_identifiers(expression: E, dialect: DialectType = None, identify: bool = True) -> E:
+    """Makes sure all identifiers that need to be quoted are quoted."""
+
+    def _quote(expression: E) -> E:
+        if isinstance(expression, exp.Identifier):
+            name = expression.this
+            expression.set(
+                "quoted",
+                identify
+                or case_sensitive(name, dialect=dialect)
+                or not exp.SAFE_IDENTIFIER_RE.match(name),
+            )
+        return expression
+
+    return expression.transform(_quote, copy=False)
+
+
 class Resolver:
     """
     Helper for resolving columns.
@@ -469,9 +485,7 @@ class Resolver:
         if node_alias:
             return exp.to_identifier(node_alias.this)
 
-        return exp.to_identifier(
-            table_name, quoted=node.this.quoted if isinstance(node, exp.Table) else None
-        )
+        return exp.to_identifier(table_name)
 
     @property
     def all_columns(self):
