@@ -51,7 +51,6 @@ class TokenType(AutoName):
     DOLLAR = auto()
     PARAMETER = auto()
     SESSION_PARAMETER = auto()
-    NATIONAL = auto()
     DAMP = auto()
 
     BLOCK_START = auto()
@@ -72,6 +71,8 @@ class TokenType(AutoName):
     BIT_STRING = auto()
     HEX_STRING = auto()
     BYTE_STRING = auto()
+    NATIONAL_STRING = auto()
+    RAW_STRING = auto()
 
     # types
     BIT = auto()
@@ -112,6 +113,18 @@ class TokenType(AutoName):
     DATETIME = auto()
     DATETIME64 = auto()
     DATE = auto()
+    INT4RANGE = auto()
+    INT4MULTIRANGE = auto()
+    INT8RANGE = auto()
+    INT8MULTIRANGE = auto()
+    NUMRANGE = auto()
+    NUMMULTIRANGE = auto()
+    TSRANGE = auto()
+    TSMULTIRANGE = auto()
+    TSTZRANGE = auto()
+    TSTZMULTIRANGE = auto()
+    DATERANGE = auto()
+    DATEMULTIRANGE = auto()
     UUID = auto()
     GEOGRAPHY = auto()
     NULLABLE = auto()
@@ -131,6 +144,7 @@ class TokenType(AutoName):
     VARIANT = auto()
     OBJECT = auto()
     INET = auto()
+    ENUM = auto()
 
     # keywords
     ALIAS = auto()
@@ -143,30 +157,22 @@ class TokenType(AutoName):
     ARRAY = auto()
     ASC = auto()
     ASOF = auto()
-    AT_TIME_ZONE = auto()
     AUTO_INCREMENT = auto()
     BEGIN = auto()
     BETWEEN = auto()
-    BOTH = auto()
-    BUCKET = auto()
-    BY_DEFAULT = auto()
     CACHE = auto()
-    CASCADE = auto()
     CASE = auto()
     CHARACTER_SET = auto()
-    CLUSTER_BY = auto()
     COLLATE = auto()
     COMMAND = auto()
     COMMENT = auto()
     COMMIT = auto()
-    COMPOUND = auto()
     CONSTRAINT = auto()
     CREATE = auto()
     CROSS = auto()
     CUBE = auto()
     CURRENT_DATE = auto()
     CURRENT_DATETIME = auto()
-    CURRENT_ROW = auto()
     CURRENT_TIME = auto()
     CURRENT_TIMESTAMP = auto()
     CURRENT_USER = auto()
@@ -174,9 +180,8 @@ class TokenType(AutoName):
     DELETE = auto()
     DESC = auto()
     DESCRIBE = auto()
+    DICTIONARY = auto()
     DISTINCT = auto()
-    DISTINCT_FROM = auto()
-    DISTRIBUTE_BY = auto()
     DIV = auto()
     DROP = auto()
     ELSE = auto()
@@ -190,7 +195,6 @@ class TokenType(AutoName):
     FILTER = auto()
     FINAL = auto()
     FIRST = auto()
-    FOLLOWING = auto()
     FOR = auto()
     FOREIGN_KEY = auto()
     FORMAT = auto()
@@ -204,7 +208,6 @@ class TokenType(AutoName):
     HAVING = auto()
     HINT = auto()
     IF = auto()
-    IGNORE_NULLS = auto()
     ILIKE = auto()
     ILIKE_ANY = auto()
     IN = auto()
@@ -223,37 +226,27 @@ class TokenType(AutoName):
     KEEP = auto()
     LANGUAGE = auto()
     LATERAL = auto()
-    LAZY = auto()
-    LEADING = auto()
     LEFT = auto()
     LIKE = auto()
     LIKE_ANY = auto()
     LIMIT = auto()
-    LOAD_DATA = auto()
-    LOCAL = auto()
+    LOAD = auto()
     LOCK = auto()
     MAP = auto()
     MATCH_RECOGNIZE = auto()
-    MATERIALIZED = auto()
     MERGE = auto()
     MOD = auto()
     NATURAL = auto()
     NEXT = auto()
     NEXT_VALUE_FOR = auto()
-    NO_ACTION = auto()
     NOTNULL = auto()
     NULL = auto()
-    NULLS_FIRST = auto()
-    NULLS_LAST = auto()
     OFFSET = auto()
     ON = auto()
-    ONLY = auto()
-    OPTIONS = auto()
     ORDER_BY = auto()
     ORDERED = auto()
     ORDINALITY = auto()
     OUTER = auto()
-    OUT_OF = auto()
     OVER = auto()
     OVERLAPS = auto()
     OVERWRITE = auto()
@@ -263,7 +256,6 @@ class TokenType(AutoName):
     PIVOT = auto()
     PLACEHOLDER = auto()
     PRAGMA = auto()
-    PRECEDING = auto()
     PRIMARY_KEY = auto()
     PROCEDURE = auto()
     PROPERTIES = auto()
@@ -273,7 +265,6 @@ class TokenType(AutoName):
     RANGE = auto()
     RECURSIVE = auto()
     REPLACE = auto()
-    RESPECT_NULLS = auto()
     RETURNING = auto()
     REFERENCES = auto()
     RIGHT = auto()
@@ -282,7 +273,6 @@ class TokenType(AutoName):
     ROLLUP = auto()
     ROW = auto()
     ROWS = auto()
-    SEED = auto()
     SELECT = auto()
     SEMI = auto()
     SEPARATOR = auto()
@@ -292,19 +282,14 @@ class TokenType(AutoName):
     SHOW = auto()
     SIMILAR_TO = auto()
     SOME = auto()
-    SORTKEY = auto()
-    SORT_BY = auto()
     STRUCT = auto()
     TABLE_SAMPLE = auto()
     TEMPORARY = auto()
     TOP = auto()
     THEN = auto()
-    TRAILING = auto()
     TRUE = auto()
-    UNBOUNDED = auto()
     UNCACHE = auto()
     UNION = auto()
-    UNLOGGED = auto()
     UNNEST = auto()
     UNPIVOT = auto()
     UPDATE = auto()
@@ -317,15 +302,11 @@ class TokenType(AutoName):
     WHERE = auto()
     WINDOW = auto()
     WITH = auto()
-    WITH_TIME_ZONE = auto()
-    WITH_LOCAL_TIME_ZONE = auto()
-    WITHIN_GROUP = auto()
-    WITHOUT_TIME_ZONE = auto()
     UNIQUE = auto()
 
 
 class Token:
-    __slots__ = ("token_type", "text", "line", "col", "end", "comments")
+    __slots__ = ("token_type", "text", "line", "col", "start", "end", "comments")
 
     @classmethod
     def number(cls, number: int) -> Token:
@@ -353,22 +334,29 @@ class Token:
         text: str,
         line: int = 1,
         col: int = 1,
+        start: int = 0,
         end: int = 0,
         comments: t.List[str] = [],
     ) -> None:
+        """Token initializer.
+
+        Args:
+            token_type: The TokenType Enum.
+            text: The text of the token.
+            line: The line that the token ends on.
+            col: The column that the token ends on.
+            start: The start index of the token.
+            end: The ending index of the token.
+            comments: The comments to attach to the token.
+        """
         self.token_type = token_type
         self.text = text
         self.line = line
-        size = len(text)
         self.col = col
-        self.end = end if end else size
+        self.start = start
+        self.end = end
         self.comments = comments
 
-    @property
-    def start(self) -> int:
-        """Returns the start of the token."""
-        return self.end - len(self.text)
-
     def __repr__(self) -> str:
         attributes = ", ".join(f"{k}: {getattr(self, k)}" for k in self.__slots__)
         return f"<Token {attributes}>"
@@ -378,41 +366,54 @@ class _Tokenizer(type):
     def __new__(cls, clsname, bases, attrs):
         klass = super().__new__(cls, clsname, bases, attrs)
 
-        klass._QUOTES = {
-            f"{prefix}{s}": e
-            for s, e in cls._delimeter_list_to_dict(klass.QUOTES).items()
-            for prefix in (("",) if s[0].isalpha() else ("", "n", "N"))
+        def _convert_quotes(arr: t.List[str | t.Tuple[str, str]]) -> t.Dict[str, str]:
+            return dict(
+                (item, item) if isinstance(item, str) else (item[0], item[1]) for item in arr
+            )
+
+        def _quotes_to_format(
+            token_type: TokenType, arr: t.List[str | t.Tuple[str, str]]
+        ) -> t.Dict[str, t.Tuple[str, TokenType]]:
+            return {k: (v, token_type) for k, v in _convert_quotes(arr).items()}
+
+        klass._QUOTES = _convert_quotes(klass.QUOTES)
+        klass._IDENTIFIERS = _convert_quotes(klass.IDENTIFIERS)
+
+        klass._FORMAT_STRINGS = {
+            **{
+                p + s: (e, TokenType.NATIONAL_STRING)
+                for s, e in klass._QUOTES.items()
+                for p in ("n", "N")
+            },
+            **_quotes_to_format(TokenType.BIT_STRING, klass.BIT_STRINGS),
+            **_quotes_to_format(TokenType.BYTE_STRING, klass.BYTE_STRINGS),
+            **_quotes_to_format(TokenType.HEX_STRING, klass.HEX_STRINGS),
+            **_quotes_to_format(TokenType.RAW_STRING, klass.RAW_STRINGS),
         }
-        klass._BIT_STRINGS = cls._delimeter_list_to_dict(klass.BIT_STRINGS)
-        klass._HEX_STRINGS = cls._delimeter_list_to_dict(klass.HEX_STRINGS)
-        klass._BYTE_STRINGS = cls._delimeter_list_to_dict(klass.BYTE_STRINGS)
-        klass._IDENTIFIERS = cls._delimeter_list_to_dict(klass.IDENTIFIERS)
+
         klass._STRING_ESCAPES = set(klass.STRING_ESCAPES)
         klass._IDENTIFIER_ESCAPES = set(klass.IDENTIFIER_ESCAPES)
-        klass._COMMENTS = dict(
-            (comment, None) if isinstance(comment, str) else (comment[0], comment[1])
-            for comment in klass.COMMENTS
-        )
+        klass._COMMENTS = {
+            **dict(
+                (comment, None) if isinstance(comment, str) else (comment[0], comment[1])
+                for comment in klass.COMMENTS
+            ),
+            "{#": "#}",  # Ensure Jinja comments are tokenized correctly in all dialects
+        }
 
-        klass.KEYWORD_TRIE = new_trie(
+        klass._KEYWORD_TRIE = new_trie(
             key.upper()
-            for key in {
-                **klass.KEYWORDS,
-                **{comment: TokenType.COMMENT for comment in klass._COMMENTS},
-                **{quote: TokenType.QUOTE for quote in klass._QUOTES},
-                **{bit_string: TokenType.BIT_STRING for bit_string in klass._BIT_STRINGS},
-                **{hex_string: TokenType.HEX_STRING for hex_string in klass._HEX_STRINGS},
-                **{byte_string: TokenType.BYTE_STRING for byte_string in klass._BYTE_STRINGS},
-            }
+            for key in (
+                *klass.KEYWORDS,
+                *klass._COMMENTS,
+                *klass._QUOTES,
+                *klass._FORMAT_STRINGS,
+            )
             if " " in key or any(single in key for single in klass.SINGLE_TOKENS)
         )
 
         return klass
 
-    @staticmethod
-    def _delimeter_list_to_dict(list: t.List[str | t.Tuple[str, str]]) -> t.Dict[str, str]:
-        return dict((item, item) if isinstance(item, str) else (item[0], item[1]) for item in list)
-
 
 class Tokenizer(metaclass=_Tokenizer):
     SINGLE_TOKENS = {
@@ -453,28 +454,29 @@ class Tokenizer(metaclass=_Tokenizer):
     BIT_STRINGS: t.List[str | t.Tuple[str, str]] = []
     BYTE_STRINGS: t.List[str | t.Tuple[str, str]] = []
     HEX_STRINGS: t.List[str | t.Tuple[str, str]] = []
+    RAW_STRINGS: t.List[str | t.Tuple[str, str]] = []
     IDENTIFIERS: t.List[str | t.Tuple[str, str]] = ['"']
     IDENTIFIER_ESCAPES = ['"']
     QUOTES: t.List[t.Tuple[str, str] | str] = ["'"]
     STRING_ESCAPES = ["'"]
     VAR_SINGLE_TOKENS: t.Set[str] = set()
 
+    # Autofilled
+    IDENTIFIERS_CAN_START_WITH_DIGIT: bool = False
+
     _COMMENTS: t.Dict[str, str] = {}
-    _BIT_STRINGS: t.Dict[str, str] = {}
-    _BYTE_STRINGS: t.Dict[str, str] = {}
-    _HEX_STRINGS: t.Dict[str, str] = {}
+    _FORMAT_STRINGS: t.Dict[str, t.Tuple[str, TokenType]] = {}
     _IDENTIFIERS: t.Dict[str, str] = {}
     _IDENTIFIER_ESCAPES: t.Set[str] = set()
     _QUOTES: t.Dict[str, str] = {}
     _STRING_ESCAPES: t.Set[str] = set()
+    _KEYWORD_TRIE: t.Dict = {}
 
-    KEYWORDS: t.Dict[t.Optional[str], TokenType] = {
+    KEYWORDS: t.Dict[str, TokenType] = {
         **{f"{{%{postfix}": TokenType.BLOCK_START for postfix in ("", "+", "-")},
         **{f"{prefix}%}}": TokenType.BLOCK_END for prefix in ("", "+", "-")},
-        "{{+": TokenType.BLOCK_START,
-        "{{-": TokenType.BLOCK_START,
-        "+}}": TokenType.BLOCK_END,
-        "-}}": TokenType.BLOCK_END,
+        **{f"{{{{{postfix}": TokenType.BLOCK_START for postfix in ("+", "-")},
+        **{f"{prefix}}}}}": TokenType.BLOCK_END for prefix in ("+", "-")},
         "/*+": TokenType.HINT,
         "==": TokenType.EQ,
         "::": TokenType.DCOLON,
@@ -498,30 +500,23 @@ class Tokenizer(metaclass=_Tokenizer):
         "ANY": TokenType.ANY,
         "ASC": TokenType.ASC,
         "AS": TokenType.ALIAS,
-        "AT TIME ZONE": TokenType.AT_TIME_ZONE,
+        "ASOF": TokenType.ASOF,
         "AUTOINCREMENT": TokenType.AUTO_INCREMENT,
         "AUTO_INCREMENT": TokenType.AUTO_INCREMENT,
         "BEGIN": TokenType.BEGIN,
         "BETWEEN": TokenType.BETWEEN,
-        "BOTH": TokenType.BOTH,
-        "BUCKET": TokenType.BUCKET,
-        "BY DEFAULT": TokenType.BY_DEFAULT,
         "CACHE": TokenType.CACHE,
         "UNCACHE": TokenType.UNCACHE,
         "CASE": TokenType.CASE,
-        "CASCADE": TokenType.CASCADE,
         "CHARACTER SET": TokenType.CHARACTER_SET,
-        "CLUSTER BY": TokenType.CLUSTER_BY,
         "COLLATE": TokenType.COLLATE,
         "COLUMN": TokenType.COLUMN,
         "COMMIT": TokenType.COMMIT,
-        "COMPOUND": TokenType.COMPOUND,
         "CONSTRAINT": TokenType.CONSTRAINT,
         "CREATE": TokenType.CREATE,
         "CROSS": TokenType.CROSS,
         "CUBE": TokenType.CUBE,
         "CURRENT_DATE": TokenType.CURRENT_DATE,
-        "CURRENT ROW": TokenType.CURRENT_ROW,
         "CURRENT_TIME": TokenType.CURRENT_TIME,
         "CURRENT_TIMESTAMP": TokenType.CURRENT_TIMESTAMP,
         "CURRENT_USER": TokenType.CURRENT_USER,
@@ -531,8 +526,6 @@ class Tokenizer(metaclass=_Tokenizer):
         "DESC": TokenType.DESC,
         "DESCRIBE": TokenType.DESCRIBE,
         "DISTINCT": TokenType.DISTINCT,
-        "DISTINCT FROM": TokenType.DISTINCT_FROM,
-        "DISTRIBUTE BY": TokenType.DISTRIBUTE_BY,
         "DIV": TokenType.DIV,
         "DROP": TokenType.DROP,
         "ELSE": TokenType.ELSE,
@@ -547,7 +540,6 @@ class Tokenizer(metaclass=_Tokenizer):
         "FIRST": TokenType.FIRST,
         "FULL": TokenType.FULL,
         "FUNCTION": TokenType.FUNCTION,
-        "FOLLOWING": TokenType.FOLLOWING,
         "FOR": TokenType.FOR,
         "FOREIGN KEY": TokenType.FOREIGN_KEY,
         "FORMAT": TokenType.FORMAT,
@@ -560,7 +552,6 @@ class Tokenizer(metaclass=_Tokenizer):
         "HAVING": TokenType.HAVING,
         "IF": TokenType.IF,
         "ILIKE": TokenType.ILIKE,
-        "IGNORE NULLS": TokenType.IGNORE_NULLS,
         "IN": TokenType.IN,
         "INDEX": TokenType.INDEX,
         "INET": TokenType.INET,
@@ -574,35 +565,25 @@ class Tokenizer(metaclass=_Tokenizer):
         "JOIN": TokenType.JOIN,
         "KEEP": TokenType.KEEP,
         "LATERAL": TokenType.LATERAL,
-        "LAZY": TokenType.LAZY,
-        "LEADING": TokenType.LEADING,
         "LEFT": TokenType.LEFT,
         "LIKE": TokenType.LIKE,
         "LIMIT": TokenType.LIMIT,
-        "LOAD DATA": TokenType.LOAD_DATA,
-        "LOCAL": TokenType.LOCAL,
+        "LOAD": TokenType.LOAD,
         "LOCK": TokenType.LOCK,
-        "MATERIALIZED": TokenType.MATERIALIZED,
         "MERGE": TokenType.MERGE,
         "NATURAL": TokenType.NATURAL,
         "NEXT": TokenType.NEXT,
         "NEXT VALUE FOR": TokenType.NEXT_VALUE_FOR,
-        "NO ACTION": TokenType.NO_ACTION,
         "NOT": TokenType.NOT,
         "NOTNULL": TokenType.NOTNULL,
         "NULL": TokenType.NULL,
-        "NULLS FIRST": TokenType.NULLS_FIRST,
-        "NULLS LAST": TokenType.NULLS_LAST,
         "OBJECT": TokenType.OBJECT,
         "OFFSET": TokenType.OFFSET,
         "ON": TokenType.ON,
-        "ONLY": TokenType.ONLY,
-        "OPTIONS": TokenType.OPTIONS,
         "OR": TokenType.OR,
         "ORDER BY": TokenType.ORDER_BY,
         "ORDINALITY": TokenType.ORDINALITY,
         "OUTER": TokenType.OUTER,
-        "OUT OF": TokenType.OUT_OF,
         "OVER": TokenType.OVER,
         "OVERLAPS": TokenType.OVERLAPS,
         "OVERWRITE": TokenType.OVERWRITE,
@@ -613,7 +594,6 @@ class Tokenizer(metaclass=_Tokenizer):
         "PERCENT": TokenType.PERCENT,
         "PIVOT": TokenType.PIVOT,
         "PRAGMA": TokenType.PRAGMA,
-        "PRECEDING": TokenType.PRECEDING,
         "PRIMARY KEY": TokenType.PRIMARY_KEY,
         "PROCEDURE": TokenType.PROCEDURE,
         "QUALIFY": TokenType.QUALIFY,
@@ -621,7 +601,7 @@ class Tokenizer(metaclass=_Tokenizer):
         "RECURSIVE": TokenType.RECURSIVE,
         "REGEXP": TokenType.RLIKE,
         "REPLACE": TokenType.REPLACE,
-        "RESPECT NULLS": TokenType.RESPECT_NULLS,
+        "RETURNING": TokenType.RETURNING,
         "REFERENCES": TokenType.REFERENCES,
         "RIGHT": TokenType.RIGHT,
         "RLIKE": TokenType.RLIKE,
@@ -630,7 +610,6 @@ class Tokenizer(metaclass=_Tokenizer):
         "ROW": TokenType.ROW,
         "ROWS": TokenType.ROWS,
         "SCHEMA": TokenType.SCHEMA,
-        "SEED": TokenType.SEED,
         "SELECT": TokenType.SELECT,
         "SEMI": TokenType.SEMI,
         "SET": TokenType.SET,
@@ -638,18 +617,13 @@ class Tokenizer(metaclass=_Tokenizer):
         "SHOW": TokenType.SHOW,
         "SIMILAR TO": TokenType.SIMILAR_TO,
         "SOME": TokenType.SOME,
-        "SORTKEY": TokenType.SORTKEY,
-        "SORT BY": TokenType.SORT_BY,
         "TABLE": TokenType.TABLE,
         "TABLESAMPLE": TokenType.TABLE_SAMPLE,
         "TEMP": TokenType.TEMPORARY,
         "TEMPORARY": TokenType.TEMPORARY,
         "THEN": TokenType.THEN,
         "TRUE": TokenType.TRUE,
-        "TRAILING": TokenType.TRAILING,
-        "UNBOUNDED": TokenType.UNBOUNDED,
         "UNION": TokenType.UNION,
-        "UNLOGGED": TokenType.UNLOGGED,
         "UNNEST": TokenType.UNNEST,
         "UNPIVOT": TokenType.UNPIVOT,
         "UPDATE": TokenType.UPDATE,
@@ -663,10 +637,6 @@ class Tokenizer(metaclass=_Tokenizer):
         "WHERE": TokenType.WHERE,
         "WINDOW": TokenType.WINDOW,
         "WITH": TokenType.WITH,
-        "WITH TIME ZONE": TokenType.WITH_TIME_ZONE,
-        "WITH LOCAL TIME ZONE": TokenType.WITH_LOCAL_TIME_ZONE,
-        "WITHIN GROUP": TokenType.WITHIN_GROUP,
-        "WITHOUT TIME ZONE": TokenType.WITHOUT_TIME_ZONE,
         "APPLY": TokenType.APPLY,
         "ARRAY": TokenType.ARRAY,
         "BIT": TokenType.BIT,
@@ -721,19 +691,22 @@ class Tokenizer(metaclass=_Tokenizer):
         "TIMESTAMPLTZ": TokenType.TIMESTAMPLTZ,
         "DATE": TokenType.DATE,
         "DATETIME": TokenType.DATETIME,
+        "INT4RANGE": TokenType.INT4RANGE,
+        "INT4MULTIRANGE": TokenType.INT4MULTIRANGE,
+        "INT8RANGE": TokenType.INT8RANGE,
+        "INT8MULTIRANGE": TokenType.INT8MULTIRANGE,
+        "NUMRANGE": TokenType.NUMRANGE,
+        "NUMMULTIRANGE": TokenType.NUMMULTIRANGE,
+        "TSRANGE": TokenType.TSRANGE,
+        "TSMULTIRANGE": TokenType.TSMULTIRANGE,
+        "TSTZRANGE": TokenType.TSTZRANGE,
+        "TSTZMULTIRANGE": TokenType.TSTZMULTIRANGE,
+        "DATERANGE": TokenType.DATERANGE,
+        "DATEMULTIRANGE": TokenType.DATEMULTIRANGE,
         "UNIQUE": TokenType.UNIQUE,
         "STRUCT": TokenType.STRUCT,
         "VARIANT": TokenType.VARIANT,
         "ALTER": TokenType.ALTER,
-        "ALTER AGGREGATE": TokenType.COMMAND,
-        "ALTER DEFAULT": TokenType.COMMAND,
-        "ALTER DOMAIN": TokenType.COMMAND,
-        "ALTER ROLE": TokenType.COMMAND,
-        "ALTER RULE": TokenType.COMMAND,
-        "ALTER SEQUENCE": TokenType.COMMAND,
-        "ALTER TYPE": TokenType.COMMAND,
-        "ALTER USER": TokenType.COMMAND,
-        "ALTER VIEW": TokenType.COMMAND,
         "ANALYZE": TokenType.COMMAND,
         "CALL": TokenType.COMMAND,
         "COMMENT": TokenType.COMMENT,
@@ -767,10 +740,7 @@ class Tokenizer(metaclass=_Tokenizer):
     NUMERIC_LITERALS: t.Dict[str, str] = {}
     ENCODE: t.Optional[str] = None
 
-    COMMENTS = ["--", ("/*", "*/"), ("{#", "#}")]
-    KEYWORD_TRIE: t.Dict = {}  # autofilled
-
-    IDENTIFIER_CAN_START_WITH_DIGIT = False
+    COMMENTS = ["--", ("/*", "*/")]
 
     __slots__ = (
         "sql",
@@ -797,7 +767,7 @@ class Tokenizer(metaclass=_Tokenizer):
         self._start = 0
         self._current = 0
         self._line = 1
-        self._col = 1
+        self._col = 0
         self._comments: t.List[str] = []
 
         self._char = ""
@@ -810,13 +780,12 @@ class Tokenizer(metaclass=_Tokenizer):
         self.reset()
         self.sql = sql
         self.size = len(sql)
+
         try:
             self._scan()
         except Exception as e:
-            start = self._current - 50
-            end = self._current + 50
-            start = start if start > 0 else 0
-            end = end if end < self.size else self.size - 1
+            start = max(self._current - 50, 0)
+            end = min(self._current + 50, self.size - 1)
             context = self.sql[start:end]
             raise ValueError(f"Error tokenizing '{context}'") from e
 
@@ -841,17 +810,17 @@ class Tokenizer(metaclass=_Tokenizer):
             if until and until():
                 break
 
-        if self.tokens:
+        if self.tokens and self._comments:
             self.tokens[-1].comments.extend(self._comments)
 
     def _chars(self, size: int) -> str:
         if size == 1:
             return self._char
+
         start = self._current - 1
         end = start + size
-        if end <= self.size:
-            return self.sql[start:end]
-        return ""
+
+        return self.sql[start:end] if end <= self.size else ""
 
     def _advance(self, i: int = 1, alnum: bool = False) -> None:
         if self.WHITE_SPACE.get(self._char) is TokenType.BREAK:
@@ -866,6 +835,7 @@ class Tokenizer(metaclass=_Tokenizer):
         self._peek = "" if self._end else self.sql[self._current]
 
         if alnum and self._char.isalnum():
+            # Here we use local variables instead of attributes for better performance
             _col = self._col
             _current = self._current
             _end = self._end
@@ -887,16 +857,23 @@ class Tokenizer(metaclass=_Tokenizer):
     def _text(self) -> str:
         return self.sql[self._start : self._current]
 
+    def peek(self, i: int = 0) -> str:
+        i = self._current + i
+        if i < self.size:
+            return self.sql[i]
+        return ""
+
     def _add(self, token_type: TokenType, text: t.Optional[str] = None) -> None:
         self._prev_token_line = self._line
         self.tokens.append(
             Token(
                 token_type,
-                self._text if text is None else text,
-                self._line,
-                self._col,
-                self._current,
-                self._comments,
+                text=self._text if text is None else text,
+                line=self._line,
+                col=self._col,
+                start=self._start,
+                end=self._current - 1,
+                comments=self._comments,
             )
         )
         self._comments = []
@@ -923,7 +900,7 @@ class Tokenizer(metaclass=_Tokenizer):
         char = chars
         prev_space = False
         skip = False
-        trie = self.KEYWORD_TRIE
+        trie = self._KEYWORD_TRIE
         single_token = char in self.SINGLE_TOKENS
 
         while chars:
@@ -968,8 +945,6 @@ class Tokenizer(metaclass=_Tokenizer):
 
         if self._scan_string(word):
             return
-        if self._scan_formatted_string(word):
-            return
         if self._scan_comment(word):
             return
 
@@ -1013,9 +988,9 @@ class Tokenizer(metaclass=_Tokenizer):
         if self._char == "0":
             peek = self._peek.upper()
             if peek == "B":
-                return self._scan_bits() if self._BIT_STRINGS else self._add(TokenType.NUMBER)
+                return self._scan_bits() if self.BIT_STRINGS else self._add(TokenType.NUMBER)
             elif peek == "X":
-                return self._scan_hex() if self._HEX_STRINGS else self._add(TokenType.NUMBER)
+                return self._scan_hex() if self.HEX_STRINGS else self._add(TokenType.NUMBER)
 
         decimal = False
         scientific = 0
@@ -1024,8 +999,12 @@ class Tokenizer(metaclass=_Tokenizer):
             if self._peek.isdigit():
                 self._advance()
             elif self._peek == "." and not decimal:
-                decimal = True
-                self._advance()
+                after = self.peek(1)
+                if after.isdigit() or not after.isalpha():
+                    decimal = True
+                    self._advance()
+                else:
+                    return self._add(TokenType.VAR)
             elif self._peek in ("-", "+") and scientific == 1:
                 scientific += 1
                 self._advance()
@@ -1040,13 +1019,13 @@ class Tokenizer(metaclass=_Tokenizer):
                     literal += self._peek.upper()
                     self._advance()
 
-                token_type = self.KEYWORDS.get(self.NUMERIC_LITERALS.get(literal))
+                token_type = self.KEYWORDS.get(self.NUMERIC_LITERALS.get(literal, ""))
 
                 if token_type:
                     self._add(TokenType.NUMBER, number_text)
                     self._add(TokenType.DCOLON, "::")
                     return self._add(token_type, literal)
-                elif self.IDENTIFIER_CAN_START_WITH_DIGIT:
+                elif self.IDENTIFIERS_CAN_START_WITH_DIGIT:
                     return self._add(TokenType.VAR)
 
                 self._add(TokenType.NUMBER, number_text)
@@ -1084,37 +1063,24 @@ class Tokenizer(metaclass=_Tokenizer):
 
         return self._text
 
-    def _scan_string(self, quote: str) -> bool:
-        quote_end = self._QUOTES.get(quote)
-        if quote_end is None:
-            return False
+    def _scan_string(self, start: str) -> bool:
+        base = None
+        token_type = TokenType.STRING
 
-        self._advance(len(quote))
-        text = self._extract_string(quote_end)
-        text = text.encode(self.ENCODE).decode(self.ENCODE) if self.ENCODE else text
-        self._add(TokenType.NATIONAL if quote[0].upper() == "N" else TokenType.STRING, text)
-        return True
+        if start in self._QUOTES:
+            end = self._QUOTES[start]
+        elif start in self._FORMAT_STRINGS:
+            end, token_type = self._FORMAT_STRINGS[start]
 
-    # X'1234', b'0110', E'\\\\\' etc.
-    def _scan_formatted_string(self, string_start: str) -> bool:
-        if string_start in self._HEX_STRINGS:
-            delimiters = self._HEX_STRINGS
-            token_type = TokenType.HEX_STRING
-            base = 16
-        elif string_start in self._BIT_STRINGS:
-            delimiters = self._BIT_STRINGS
-            token_type = TokenType.BIT_STRING
-            base = 2
-        elif string_start in self._BYTE_STRINGS:
-            delimiters = self._BYTE_STRINGS
-            token_type = TokenType.BYTE_STRING
-            base = None
+            if token_type == TokenType.HEX_STRING:
+                base = 16
+            elif token_type == TokenType.BIT_STRING:
+                base = 2
         else:
             return False
 
-        self._advance(len(string_start))
-        string_end = delimiters[string_start]
-        text = self._extract_string(string_end)
+        self._advance(len(start))
+        text = self._extract_string(end)
 
         if base:
             try:
@@ -1123,6 +1089,8 @@ class Tokenizer(metaclass=_Tokenizer):
                 raise RuntimeError(
                     f"Numeric string contains invalid characters from {self._line}:{self._start}"
                 )
+        else:
+            text = text.encode(self.ENCODE).decode(self.ENCODE) if self.ENCODE else text
 
         self._add(token_type, text)
         return True
